/*
 * @(#)TIC.java	1.1 97/03/24
 *
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/**
 * Tiny Interpreter Command-loop.
 * <p>
 * (Compare the neurological scope and complexity of a Tic with
 * that of a full-blown Thought or Scheme.)
 * <p>
 * This is a demonstration of the Java Core Reflection API.
 * As such, the heart of the code is the methods on the second page,
 * after the comment "REFLECTIVE CODE STARTS HERE".
 * This hack can clearly be corrected, extended, and solidified in many ways.
 * <p>
 * Example session:

 * <pre>
 &gt; System.out.println "hello, world"
 hello, world
 &gt; { 1 2 }
 $1 = { 1 2 }
 &gt; java.lang.reflect.Array.get  {"zero" 1 2.0}  2
 $2 = 2
 &gt; (new StringBuffer)
 $3 = 
 &gt; $3.append "yow"
 $4 = yow
 &gt; System.out.println ($3.append "sah")
 yowsah
 &gt; new Frame "Foo"
 $5 = java.awt.Frame[frame0,0,0,0x0,invalid,hidden,layout=java.awt.BorderLayout,resizable,title=Foo]
 &gt; $5.setSize 200 300
 &gt; $5.show
 * </pre>

 * @author John R. Rose
 */

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

public class TIC {
  protected Reader charSource;
  protected StreamTokenizer in;
  protected int historyIndex = 0;
  protected Hashtable names = new Hashtable();
  {
    names.put("true", Boolean.TRUE);
    names.put("false", Boolean.FALSE);
    names.put("null", new Forceable() {
      public Object force() { return null; }
    });
    names.put("new", new Forceable() {
      public Object apply(Object args[]) throws Exception {
	if (args.length == 0)  error("new: class name required");
	return construct((Class)args[0], shift(args));
      }
    });
  }
  protected String imports[] = {
    "java.lang", "java.util", "java.io", "java.awt"
  };

  public static void main(String av[]) throws IOException {
    new TIC(new InputStreamReader(System.in)).commandLoop();
  }
  public TIC(Reader charSource) {
    this.charSource = charSource;
    in = new StreamTokenizer(charSource);
    in.eolIsSignificant(true);
    String wc = "$_.";		// other name constituents
    for (int i = 0; i < wc.length(); i++)
      in.wordChars(wc.charAt(i), wc.charAt(i));
  }
  public void commandLoop() throws IOException {
    while (true) {
      if (!charSource.ready())  System.err.print("> ");
      in.nextToken();		// prime the pump
      if (in.ttype == StreamTokenizer.TT_EOF)  break;
      Object x = this;
      try {
	Object words[] = readArray(StreamTokenizer.TT_EOL);
	if (words.length > 1)
	  x = new List(words[0], shift(words));
	else
	  x = (words.length == 0) ? null : words[0];
	x = force(eval(x));
      } catch (Exception ee) {
	System.out.println(ee);
	if (x == this && ee instanceof IOException)  throw (IOException)ee;
	while (charSource.ready())  charSource.read();
	continue;
      }

      if (x == null)  continue;
      String historyName = "$" + (++historyIndex);
      names.put(historyName, x);
      System.out.println(historyName + " = " + printString(x));

      if (!(x.getClass().isArray() ||
	    x instanceof String || x instanceof Number ||
	    x instanceof Boolean || x instanceof Character))
	names.put("$", x);	// last message recipient
    }
    System.out.println();
  }
  public String printString(Object x) {
    if (x.getClass().isArray()) {
	StringBuffer b = new StringBuffer("{");
	for (int i = 0; i < Array.getLength(x); i++)
	  b.append(" ").append(printString(Array.get(x, i)));
	return b.append(" }").toString();
    }
    if (x instanceof String)  return "\"" + x + "\"";
    return x.toString();
  }

  public Object[] readArray(int until) throws IOException {
    Vector words = new Vector();
    loop: while (true) {
      Object word;
      switch (in.ttype) {
      case StreamTokenizer.TT_NUMBER:
	if (in.nval == (int)in.nval)
	  word = new Integer((int)in.nval);
	else if (in.nval == (float)in.nval)
	  word = new Float((float)in.nval);
	else
	  word = new Double(in.nval);
	break;
      case '\'':
	if (in.sval.length() == 1) {
	  word = new Character(in.sval.charAt(0));
	  break;
	} //else it's a string:
      case '"':
	word = in.sval;
	break;
      case StreamTokenizer.TT_WORD:
	word = new Word(in.sval);
	break;
      case StreamTokenizer.TT_EOL:
	if (until == in.ttype)  break loop;
	in.nextToken();
	continue;
      case StreamTokenizer.TT_EOF:
	break loop;
      default:
	if (until == in.ttype)  break loop;
	error("bad char " + (char)in.ttype); return null;
      case '{':
	in.nextToken();
	final Object elements[] = readArray('}'); 
	word = new Node() {
	  public Object eval() throws Exception {
	    return evalArray(elements);
	  }
	};
	break;
      case '(':
	in.nextToken();
	Object subWords[] = readArray(')'); 
	word = new List(subWords[0], shift(subWords));
	break;
      }
      words.addElement(word);
      in.nextToken();
    }
    Object a[] = new Object[words.size()];
    words.copyInto(a);
    return a;
  }

  Object[] shift(Object a[]) {
    Object a1[] = new Object[a.length - 1];
    System.arraycopy(a, 1, a1, 0, a1.length);
    return a1;
  }

  public Object eval(Object x) throws Exception {
    return (x instanceof Node) ? ((Node)x).eval() : x;
  }
  public Object force(Object x) throws Exception {
    return (x instanceof Forceable) ? ((Forceable) x).force() : x;
  }
  public Object apply(Object x, Object args[]) throws Exception {
    return (x instanceof Forceable) ? ((Forceable)x).apply(args) :
	   error("not a method or operator: " + x);
  }
  public Object error(String message) {
    throw new RuntimeException(message);
  }
  public Object[] evalArray(Object args[]) throws Exception {
    args = (Object[]) args.clone();
    for (int i = 0; i < args.length; i++)
      args[i] = force(eval(args[i]));
    return args;
  }

  public abstract class Forceable {  // an incomplete evaluation
    public Object force() throws Exception {
      return apply(new Object[0]);
    }
    public Object apply(Object args[]) throws Exception {
      if (args.length == 0)  return force();
      return error("not a method or operator: " + this);
    }
  }
  public abstract class Node extends Forceable {
    public Object force() throws Exception {
      return error("incomplete expression: " + this);
    }
    public abstract Object eval() throws Exception;
  }
  public class Word extends Node {
    final public Object qual;	// stuff before the dot, if any
    final public String name;

    public String toString() {	// return a qualified name
      return (qual == null) ? name : qual + "." + name;
    }
    public Word(String name) {
      Object qual = null;
      int dot = name.lastIndexOf('.');
      if (dot >= 0) {
	qual = new Word(dot == 0 ? "$" : name.substring(0, dot));
	name = name.substring(dot + 1);
      }
      this.qual = qual; this.name = name;
      if (name.equals(""))  error("name missing: " + this);
    }
    public Object eval() throws Exception {
      if (qual == null) {	// a simple name
	Object x = names.get(name);
	if (x == null)  x = findClass(name);
	for (int i = 0; i < imports.length; i++) {
	  if (x != null)  break;
	  x = findClass(imports[i] + "." + name);
	}
	if (x == null)  x = this; // might be a package prefix
	return x;
      }
      Object qual = TIC.this.eval(this.qual);
      if (qual instanceof Word) { // check for package prefix
	Class clazz = findClass("" + this);
	return clazz != null ? (Object) clazz : (Object) this;
      }
      final Object target = TIC.this.force(qual);
      return new Forceable() {
	public Object force() throws Exception {
	  return select(target, name);
	}
	public Object apply(Object args[]) throws Exception {
	  return invoke(target, name, args);
	}
      };
    }
  }
  public class List extends Node {
    final public Object head;   // should eval to a forceable
    final public Object args[];

    public List(Object head, Object args[]) {
      this.head = head; this.args = args;
    }
    public Object eval() throws Exception {
      return TIC.this.apply(TIC.this.eval(head), evalArray(args));
    }
  }


  //// REFLECTIVE CODE STARTS HERE ////

  public Class findClass(String name) {
    try {
      return Class.forName(name);
    } catch (ClassNotFoundException ee) {
      return null;
    }
  }

  private Class wraptab[] = {
    Short.class, Short.TYPE,
    Byte.class, Byte.TYPE,
    Long.class, Long.TYPE,
    Float.class, Float.TYPE,
    Double.class, Double.TYPE,
    Boolean.class, Boolean.TYPE,
    Character.class, Character.TYPE,
    Integer.class, Integer.TYPE,
  };
  private Class toPrim(Class wrapper) {
    for (int i = wraptab.length - 2; i >= 0; i -= 2)
      if (wrapper == wraptab[i])
	return wraptab[i + 1];
    return wrapper;
  }

  private final Class ANYREF = null;

  public Class[] getArgumentTypes(Object args[]) {
    Class at[] = new Class[args.length];
    for (int i = 0; i < at.length; i++)
      at[i] = args[i] == null ? ANYREF : args[i].getClass();
    return at;
  }

  public boolean isMoreSpecific(Class spec[], Class gen[]) {
    if (spec.length != gen.length)  return false;
    for (int i = 0; i < spec.length; i++)
      if (spec[i] == ANYREF ? gen[i].isPrimitive()
	    : !toPrim(gen[i]).isAssignableFrom(toPrim(spec[i])))
	return false;
    return true;
  }

  public Object construct(Class clazz, Object args[])
    throws Exception {
      Class at[] = getArgumentTypes(args);
      Constructor mems[] = clazz.getConstructors();
      Constructor best = null;
      for (int i = 0; i < mems.length; i++) {
	Class pt[] = mems[i].getParameterTypes();
	if (isMoreSpecific(at, pt)) {
	  if (best == null ||
	      isMoreSpecific(pt, best.getParameterTypes()))
	    best = mems[i];
	}
      }
      if (best == null)  error("no applicable constructor for " +
			       printString(at) + ": " + clazz);
      try {
	  return best.newInstance(args);
      } catch (InvocationTargetException ee) {
	  throw (Exception) ee.getTargetException();
      }
  }

  public Object select(Object target, String name)
    throws Exception {
      Class clazz = target instanceof Class ? (Class)target
			: target.getClass();
      try {
	return clazz.getField(name).get(target);
      } catch (NoSuchFieldException ee) {
      }
      try {
	return clazz.getMethod(name, new Class[0])
	  .invoke(target, new Object[0]);
      } catch (NoSuchMethodException ee) {
      } catch (InvocationTargetException ee) {
	  throw (Exception) ee.getTargetException();
      }
      return error("no field named " + name + ": " + clazz);
  }

  public Object invoke(Object target, String name, Object args[])
    throws Exception {
      Class at[] = getArgumentTypes(args);
      Class clazz = target instanceof Class ? (Class)target
			: target.getClass();
      Method mems[] = clazz.getMethods();
      Method best = null;
      for (int i = 0; i < mems.length; i++) {
	if (!mems[i].getName().equals(name))  continue;
	Class pt[] = mems[i].getParameterTypes();
	if (isMoreSpecific(at, pt)) {
	  if (best == null ||
	      isMoreSpecific(pt, best.getParameterTypes()))
	    best = mems[i];
	}
      }
      if (best == null)  error("no applicable method " + name +
			       printString(at) + ": "+clazz);
      try {
	  return best.invoke(target, args);
      } catch (InvocationTargetException ee) {
	  throw (Exception) ee.getTargetException();
      }
  }
}
